from copy import copy

import pytest

from kks.ejudge import Deadlines, ProblemInfo, StandingsRow, Status, TaskScore
from kks.cmd.top import recalc_task_score


class FrozenProblemInfo(ProblemInfo):
    def __init__(self, past_deadline, full_score, run_penalty, current_penalty):
        super().__init__(full_score, run_penalty, current_penalty, Deadlines(None, None))
        self._past_deadline = past_deadline

    def past_deadline(self):
        return self._past_deadline


def contest_name(is_kr):
    return 'kr123' if is_kr else 'sm123'


# ---------- SCORE GENERATORS ----------


def score_ok():
    return TaskScore(contest_name(False), '100', Status.OK)


def score_review():
    return TaskScore(contest_name(False), '100', Status.REVIEW)


def score_partial():
    return TaskScore(contest_name(False), '0', Status.PARTIAL)


def score_testing(has_partial=False, is_kr=False):
    return TaskScore(contest_name(is_kr), '0' if has_partial else None, Status.TESTING)


def score_rejected(has_partial=False):
    return TaskScore(contest_name(False), '0' if has_partial else None, Status.REJECTED)


def score_not_submitted(is_kr=False):
    return TaskScore(contest_name(is_kr), None, Status.NOT_SUBMITTED)


# ---------- PROBLEM INFO GENERATORS ----------


def problem_past_deadline(full_score=100, run_penalty=10, current_penalty=100):
    return FrozenProblemInfo(True, full_score, run_penalty, current_penalty)


def normal_problem(full_score=100, run_penalty=10, current_penalty=50):
    return FrozenProblemInfo(False, full_score, run_penalty, current_penalty)


# ---------- TESTING ----------


class BaseRecalcTest:
    @pytest.fixture(autouse=True)
    def create_row(self):
        self._row = StandingsRow(1, 'user', [], 0, 0, False)

    def recalc_task_score(self, task_score, problem_info):
        recalc_task_score(self._row, task_score, problem_info)

    def assert_unchanged(self, task_score, problem_info):
        orig_score = copy(task_score)
        self.recalc_task_score(task_score, problem_info)
        assert orig_score.score == task_score.score
        assert orig_score.status == task_score.status

    def assert_updated(self, task_score, problem_info, expected_score):
        self.recalc_task_score(task_score, problem_info)
        assert task_score.score == str(expected_score)
        assert task_score.status == Status.REVIEW


class TestNormal(BaseRecalcTest):
    # Status.TESTING and past_deadline() - should never happen

    @pytest.mark.parametrize('problem_info_gen', [problem_past_deadline, normal_problem])
    @pytest.mark.parametrize('task_score_gen', [score_ok, score_review])
    def test_solved(self, problem_info_gen, task_score_gen):
        """already solved - shouldn't recalculate"""
        self.assert_unchanged(task_score_gen(), problem_info_gen())

    @pytest.mark.parametrize('task_score_gen', [score_partial, score_not_submitted])
    def test_frozen(self, task_score_gen):
        """will never be solved - shouldn't recalculate"""
        self.assert_unchanged(task_score_gen(), problem_past_deadline())

    def test_partial(self):
        task_score = score_partial()
        problem_info = normal_problem(full_score=100, run_penalty=10, current_penalty=50)
        self.assert_updated(task_score, problem_info, 40)

    def test_not_submitted(self):
        task_score = score_not_submitted()
        problem_info = normal_problem(full_score=100, current_penalty=50)
        self.assert_updated(task_score, problem_info, 50)

    @pytest.mark.parametrize('has_partial', [False, True])
    def test_testing(self, has_partial):
        task_score = score_testing(has_partial)
        problem_info = normal_problem(full_score=100, run_penalty=10, current_penalty=50)
        if has_partial:
            self.assert_updated(task_score, problem_info, 40)
        else:
            self.assert_updated(task_score, problem_info, 50)

    @pytest.mark.parametrize('problem_info_gen', [problem_past_deadline, normal_problem])
    @pytest.mark.parametrize('has_partial', [False, True])
    def test_rejected(self, problem_info_gen, has_partial):
        task_score = score_rejected(has_partial)
        problem_info = problem_info_gen(full_score=100, run_penalty=10)
        if has_partial:
            self.assert_updated(task_score, problem_info, 90)
        else:
            self.assert_updated(task_score, problem_info, 100)

    @pytest.mark.parametrize('task_score_gen', [score_partial, score_not_submitted])
    def test_min_score(self, task_score_gen):
        task_score = task_score_gen()
        problem_info = normal_problem(full_score=100, run_penalty=10, current_penalty=300)
        self.assert_updated(task_score, problem_info, 20)


class TestKr(BaseRecalcTest):
    class MockConfig:
        class options:
            max_kr = None

    @pytest.fixture(autouse=True)
    def mock_config(self, monkeypatch):
        monkeypatch.setattr('kks.cmd.top.Config', TestKr.MockConfig)

    def test_testing_maxkr(self):
        TestKr.MockConfig.options.max_kr = True
        task_score = score_testing(is_kr=True)
        self.assert_updated(task_score, problem_past_deadline(), 200)

    def test_testing_nomaxkr(self):
        TestKr.MockConfig.options.max_kr = False
        task_score = score_testing(is_kr=True)
        self.assert_unchanged(task_score, problem_past_deadline())

    @pytest.mark.parametrize('max_kr', [True, False])
    @pytest.mark.parametrize('task_score_gen', [score_testing, score_not_submitted])
    def test_running_kr(self, max_kr, task_score_gen):
        TestKr.MockConfig.options.max_kr = max_kr
        task_score = task_score_gen(is_kr=True)
        self.assert_unchanged(task_score, normal_problem())



import re

import click

from kks.ejudge import ejudge_standings
from kks.errors import AuthError, EjudgeUnavailableError
from kks.util.ejudge import EjudgeSession
from kks.util.stat import get_global_standings


@click.command(short_help='Calculate your score from homeworks')
@click.option('-b', '--barrier', 'K', type=int, default=3500,
              help='K_i param. See http://wiki.cs.hse.ru/CAOS-2022#.D0.A4.D0.BE.D1.80.D0.BC.D1.83.D0.BB.D0.B0_.D0.BE.D1.86.D0.B5.D0.BD.D0.BA.D0.B8')
@click.option('-y', '--year', type=int, default=2022,
              help='Show standings for the selected year')
@click.option('-f', '--first-contest', type=str,
              help='Show score for all contests since the chosen one (module 4 begins with sm11)')
def my_score(K, year, first_contest):
    user = None
    try:
        session = EjudgeSession()
        standings = ejudge_standings(session)
        # TODO call send_standings if opted in
        user = standings.user
    except (EjudgeUnavailableError, AuthError) as err:
        click.secho(
            f'Cannot get standings from ejudge. Reason: {err.message}', fg='yellow', err=True
        )
        return
    standings = get_global_standings(user, year)
    top1_score = get_top1_score(standings, year, first_contest)
    my_score = get_my_score(standings, year, first_contest)
    real_score = None
    if my_score < K:
        real_score = 6.0 * (my_score / K)
    else:
        real_score = min(9.0, 6.0 + 3.0 * (my_score - K) / (top1_score - K))
    fg = None
    if real_score < 4.0:
        fg = 'red'
    elif real_score < 8.0:
        fg = 'yellow'
    else:
        fg = 'green'
    print('Your current score from homework is: ', click.style(
        str(my_score), fg=fg, bold=True))
    print('Your estimated mark from homework is: ', click.style(
        f'{real_score:.02f}/9.0', fg=fg, bold=True))


def get_top1_score(standings, year, first_contest):
    contests = None
    if first_contest:
        contests = select_contests_starting_from_specific_contest(
            standings, year, first_contest)

    scores = []
    for row in standings.rows:
        score = 0
        for task in row.tasks:
            if is_rated_contest(task.contest) and task.score:
                if first_contest is None or task.contest in contests:
                    score += int(task.score)
        scores.append(score)
    return max(scores)


def get_my_score(standings, year, first_contest):
    contests = None
    if first_contest:
        contests = select_contests_starting_from_specific_contest(standings, year, first_contest)

    for row in standings.rows:
        if row.is_self:
            if contests:
                return sum(int(task.score) for task in row.tasks if task.contest in contests and task.score)
            else:
                return row.score

    return 0


def is_rated_contest(contest: str) -> bool:
    # assuming there can be split contests, like sm12.3-1, sm12_3-1, ...
    return re.match(r'^(sm|kr|ku)\d', contest) is not None


def select_contests_starting_from_specific_contest(standings, year, first_contest):
    contests = []
    for contest in standings.contests[::-1]:
        # В 2022 kr04 шла после контестов за 4 модуль, но относилась к 3 модулю.
        if not (contest.startswith('exam') or ((contest == 'kr04' or contest == 'ku04') and year == 2022)):
            contests.append(contest)
            if contest == first_contest:
                break

    return set(contests)



from enum import Enum
from mimetypes import guess_extension
from shutil import rmtree

import click

from kks.ejudge import Status, ejudge_summary, ejudge_submissions, ejudge_report
from kks.util.click import OptFlagCommand, FlagOption, OptFlagOption, Choice2
from kks.util.common import find_workspace, get_task_dir, write_contests, parse_content_type
from kks.util.ejudge import EjudgeSession
from kks.util.storage import Config


class CodeSync(Enum):
    ALL = 'all'
    REJECTS = 'rejects'


def save_needed(problem, submissions, sub_dir, session, mode: CodeSync):
    def prefix(submission):
        return f'{submission.id:05d}'

    def format_stem(submission):
        return f'{prefix(submission)}-{submission.short_status()}'

    def get_extension(submission, resp):
        mimetype, _ = parse_content_type(resp.headers.get('Content-Type', ''))
        mimetype = mimetype.lower()
        if mimetype == 'text/plain':
            return submission.suffix()
        elif mimetype == 'application/x-gzip':
            return '.gz'
        else:
            suf = guess_extension(mimetype)
            if suf is None or suf == '.':
                return ''
            return suf

    if mode == CodeSync.ALL:
        needed = submissions
    elif mode == CodeSync.REJECTS:
        needed = [sub for sub in submissions if sub.status == Status.REJECTED]
    else:
        ok, review, reject, partial = [], [], [], []
        for sub in submissions:
            if sub.status == Status.OK:
                ok.append(sub)
            elif sub.status == Status.REVIEW:
                review.append(sub)
            elif sub.status == Status.REJECTED:
                reject.append(sub)
            elif sub.status != Status.IGNORED:
                partial.append(sub)

        needed = {submissions[0]}  # always save latest
        # add last solution that passed all tests
        if ok:
            needed.add(ok[0])
        elif review:
            needed.add(review[0])
        elif reject:
            needed.add(reject[0])

    for sub in needed:
        old = list(sub_dir.glob(f'{prefix(sub)}-*'))
        file = sub_dir / format_stem(sub)
        if len(old) == 1:
            # status may change
            if old[0].stem == file.stem:
                continue
            old[0].rename(file.with_suffix(old[0].suffix))

            # may need to load the report (PR -> RJ: comments, running -> partial: failed tests)
            if sub.status not in [Status.PARTIAL, Status.REJECTED]:
                # Comments for existing PR->OK, PR->IG submissions are not synced.
                # Also, comments without status change are not updated (see stem check above)
                # TODO check new comments on Clars page?
                continue

        resp = session.get(sub.source)

        file = file.with_suffix(get_extension(sub, resp))

        source = resp.content
        if sub.status in [Status.PARTIAL, Status.REJECTED]:
            report = ejudge_report(sub.report, session)
            source = report.as_comment().encode() + source

        with open(file, 'wb') as f:
            f.write(source)


def sync_code(problem, task_dir, submissions, session, mode: CodeSync):
    sub_dir = task_dir / 'submissions'
    if sub_dir.exists():
        if not sub_dir.is_dir():
            click.secho(f'File {sub_dir.relative_to(find_workspace())} exists, skipping',
                        fg='red', err=True)
            return
    else:
        sub_dir.mkdir(parents=True, exist_ok=True)
    problem_subs = submissions.get(problem.short_name, [])
    if problem_subs:
        save_needed(problem, problem_subs, sub_dir, session, mode)


def sync_attachments(problem, dest_dir, session):
    attachments = problem.attachments()
    dest_dir_exists = dest_dir.exists()
    if dest_dir_exists and not dest_dir.is_dir():
        if attachments:
            click.secho(
                f'File {dest_dir.relative_to(find_workspace())} exists, '
                'skipping attachment sync',
                fg='red', err=True
            )
        # Don't warn if the attachment list is empty
        return

    if dest_dir_exists:
        # If an attachment was removed from the statement, it should be deleted.
        # Remove all attachments, existing ones will be resynced anyway.
        rmtree(dest_dir)

    if not attachments:
        return

    dest_dir.mkdir(parents=True, exist_ok=True)
    for att_name, url in attachments.items():
        att_path = dest_dir / att_name
        page = session.get(url)
        with att_path.open('wb') as f:
            f.write(page.content)


@click.command(short_help='Parse problems from ejudge', cls=OptFlagCommand)
@click.option('--code', cls=FlagOption, is_flag=True,
              help='Download latest submitted solutions')
@click.option('--code_opt', cls=OptFlagOption, type=Choice2(['all', 'rejects']),
              help='Download all / all rejected submissions')
@click.option('-f', '--force', is_flag=True, default=False,
              help='Force sync existing tasks')
@click.argument('filters', nargs=-1)
def sync(code, code_opt, force, filters):
    """
    Parse problems from ejudge

    If any FILTERS are specified, sync only tasks with matching prefixes/names
    """

    workspace = find_workspace()

    if workspace is None:
        click.secho(
            'You have to run sync under kks workspace (use "kks init" to create one)',
            fg='red', err=True
        )
        return

    config = Config()

    session = EjudgeSession()
    problems = ejudge_summary(session)

    code_sync_mode = CodeSync(code_opt) if code_opt else None
    code = code or code_sync_mode
    if code:
        submissions = ejudge_submissions(session)

    md_width = config.options.mdwidth

    contests = set()
    bad_contests = set()
    total_problems = 0
    old_problems = 0
    new_problems = 0

    for problem in problems:
        contest, number = problem.short_name.split('-')
        contests.add(contest)
        if filters and not any(problem.short_name.startswith(f) for f in filters):
            continue
        total_problems += 1
        if contest in bad_contests:
            continue
        task_dir = get_task_dir(workspace, contest, number)
        contest_dir = task_dir.parent

        if contest_dir.exists() and not contest_dir.is_dir():
            click.secho(f'File {contest_dir.relative_to(workspace)} exists, skipping',
                        fg='red', err=True)
            bad_contests.add(contest)
            continue

        if not task_dir.exists():
            click.secho(
                'Creating directories for task ' + click.style(problem.name, fg='blue', bold=True)
            )
            task_dir.mkdir(parents=True, exist_ok=True)
        else:
            if not task_dir.is_dir():
                click.secho(
                    f'File {task_dir.relative_to(workspace)} exists, skipping',
                    fg='red', err=True
                )
                continue

            if not force:
                if code:
                    click.secho(
                        'Syncing submissions for ' +
                        click.style(problem.name, fg='blue', bold=True)
                    )
                    sync_code(problem, task_dir, submissions, session, code_sync_mode)
                    new_problems += 1
                else:
                    old_problems += 1
                continue

            click.secho('Resyncing task ' + click.style(problem.name, fg='blue', bold=True))

        new_problems += 1

        problem = problem.get_full(session)

        main = (task_dir / problem.short_name).with_suffix(problem.suffix())
        main.touch()

        gen = task_dir / 'gen.py'

        if not gen.exists():
            with gen.open('w') as file:
                file.write('import sys\n'
                           'import random\n'
                           '\n'
                           't = int(sys.argv[1])\n'
                           'random.seed(t)\n')

        solve = task_dir / 'solve.py'
        solve.touch()

        tests_dir = task_dir / 'tests'
        tests_dir.mkdir(exist_ok=True)

        html_statement_path = task_dir / 'statement.html'
        md_statement_path = task_dir / 'statement.md'
        attachments_path = task_dir / 'attachments'

        if config.options.save_html_statements:
            # overwrite only if statement is available
            if problem.statement_available() or not html_statement_path.exists():
                with html_statement_path.open('w') as f:
                    f.write(problem.html())
        elif html_statement_path.is_file():
            html_statement_path.unlink()

        if config.options.save_md_statements:
            if problem.statement_available() or not md_statement_path.exists():
                with md_statement_path.open('w') as f:
                    f.write(problem.markdown(width=md_width))
        elif md_statement_path.is_file():
            md_statement_path.unlink()

        if config.options.save_attachments:
            if problem.statement_available():
                sync_attachments(problem, attachments_path, session)
        elif attachments_path.is_dir():
            rmtree(attachments_path)

        if problem.input_data is not None:
            with (tests_dir / '000.in').open('w') as f:
                f.write(problem.input_data)

        if problem.output_data is not None:
            with (tests_dir / '000.out').open('w') as f:
                f.write(problem.output_data)

        if code:
            click.secho('Syncing submissions')
            sync_code(problem, task_dir, submissions, session, code_sync_mode)

    write_contests(workspace, contests)

    color = 'green' if old_problems + new_problems == total_problems else 'red'
    click.secho('Sync done!', fg='green')
    click.secho(
        f'Synced tasks: {old_problems+new_problems}/{total_problems} ({old_problems} unchanged)',
        fg=color
    )



import gzip
import pickle
import sys
from configparser import ConfigParser
from datetime import datetime, timedelta
from io import BytesIO
from os import environ
from time import time
from typing import Optional, Union

import click

from kks.util.common import Singleton, config_directory


class Section:
    """
    in config.ini all option names are lowercase and all underscores are replaced with dashes
    """
    def __init__(self, config, name):
        super().__setattr__('_config', config)
        super().__setattr__('_name', Section.canonical_name(name))

    @staticmethod
    def canonical_name(name):
        return name.capitalize()

    @staticmethod
    def to_option(name):
        return name.lower().replace('_', '-')

    def _convert(self, key, value):
        type_ = self.__annotations__[key]
        if type_ is bool:
            return self._config._convert_to_boolean(value)
        return type_(value)

    def _is_option(self, key):
        return key in super().__getattribute__('__annotations__')

    def _check_key(self, key):
        if not self._is_option(key):
            raise AttributeError(f'Option "{key}" is not allowed in config section "{self._name}"')

    def __getattribute__(self, key):
        if not super().__getattribute__('_is_option')(key):
            return super().__getattribute__(key)

        value = self._config.get(self._name, Section.to_option(key), fallback=None)
        if value is not None:
            return self._convert(key, value)
        return getattr(type(self), key, None)  # default

    def __setattr__(self, key, value):
        self._check_key(key)
        if not self._config.has_section(self._name):
            self._config.add_section(self._name)
        self._config.set(self._name, Section.to_option(key), str(value))

    def __delattr__(self, key):
        self._check_key(key)
        if not self._config.has_section(self._name):
            return
        self._config.remove_option(self._name, Section.to_option(key))

    def asdict(self):
        return {key: self.__getattribute__(key) for key in super().__getattribute__('__annotations__')}

    def update(self, data: dict):
        for key, value in data.items():
            setattr(self, key, value)


class EnvSection(Section):
    @staticmethod
    def to_envvar(name):
        return name.upper()

    def __getattribute__(self, key):
        if not super().__getattribute__('_is_option')(key):
            return super(Section, self).__getattribute__(key)

        envvar = environ.get(EnvSection.to_envvar(key))
        if envvar is not None:
            return self._convert(key, envvar)
        return super().__getattribute__(key)


class AuthSection(Section):
    login: str
    password: str
    contest: int


class OptionsSection(EnvSection):
    save_html_statements: bool = True
    save_md_statements: bool = True
    save_attachments: bool = True
    mdwidth: int = 100
    max_kr: bool = False
    deadline_warning_days: int = 1
    sort_todo_by_deadline: bool = True
    global_opt_out: bool
    keep_bad_credentials: bool


class ConfigModel:
    auth: AuthSection
    options: OptionsSection


class Config(metaclass=Singleton):
    """global kks config"""

    def __init__(self):
        self._file = config_directory() / 'config.ini'
        self._config = ConfigParser()
        if self._file.is_file():
            self._config.read(self._file)

    def save(self):
        with self._file.open('w') as f:
            self._config.write(f)

    def reload(self):
        """force reload from disk"""
        if self._file.is_file():
            self._config.read(self._file)
        else:
            self._config.clear()

    def __getattribute__(self, key):
        if key in ConfigModel.__annotations__:
            section_type = ConfigModel.__annotations__[key]
            return section_type(self._config, key)
        return super().__getattribute__(key)

    def __delattr__(self, key):
        if key in ConfigModel.__annotations__:
            self._config.remove_section(Section.canonical_name(key))
        else:
            super().__delattr__(key)


class CompatUnpickler(pickle.Unpickler):

    class UniversalClass:
        def __init__(self, *args, **kwargs):
            pass

    def find_class(self, module, name):
        return CompatUnpickler.UniversalClass


class PickleStorage:
    _service_keys = ('__version__',)

    def __init__(self, name, *, compress=False, version=1):
        self.name = name
        self.compress = compress

        suffix = '.pickle.gz' if compress else '.pickle'
        self._file = (config_directory() / name).with_suffix(suffix)
        self._version = version
        self._init_data()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.save()

    def _init_data(self, clean=True):
        self._data = {'__version__': self._version}
        self._clean = clean

    def load(self):
        self._init_data()

        if self._file.exists():
            pickled_data_available = False
            try:
                with self._file.open('rb') as f:
                    data = f.read()
                if self.compress:
                    data = gzip.decompress(data)
                pickled_data_available = True
                self._data.update(pickle.loads(data))
            except Exception:
                if not (pickled_data_available and self._try_load_old(data)):
                    click.secho(f'Storage file {self._file} is corrupted', bg='red', err=True)
                    if click.confirm(click.style('Erase all saved data?', fg='red', bold=True)):
                        self._init_data(False)
                    else:
                        click.secho(
                            f'You need to fix or delete {self._file.absolute()} manually',
                            fg='red', err=True
                        )
                        sys.exit()

        if self._data['__version__'] != self._version:
            click.secho(
                f'{self._file} uses an incompatible storage version, clearing saved data',
                bg='red', err=True
            )
            self._init_data(False)
        return self

    def _try_load_old(self, encoded):
        try:
            data = CompatUnpickler(BytesIO(encoded)).load()
            if isinstance(data, dict) and data['__version__'] != self._version:
                # just an old version, we may reset the storage
                self._data = data
                return True
            return False  # something is broken

        except Exception as e:
            return False

    def keys(self):
        for k in self._data.keys():
            if k not in self._service_keys:
                yield k

    def get(self, key, default=None):
        return self._data.get(key, default)

    def set(self, key, value):
        self._data[key] = value
        self._clean = False

    def erase(self, key):
        # If key is not found, no errors are raised
        if key in self._data:
            del self._data[key]
            self._clean = False

    def clear(self):
        if len(self._data) > len(self._service_keys):
            self._init_data(False)

    def save(self):
        if self._clean:
            return
        with self._file.open('wb') as f:
            data = pickle.dumps(self._data)
            if self.compress:
                data = gzip.compress(data)
            f.write(data)
            self._clean = True

    def __del__(self):
        self.save()


class Cache(PickleStorage):
    """Storage for data with expiration times"""

    def get(self, key, default=None):
        # default is returned if value is not found or outdated
        value, exp_time = super().get(key, (None, -1))
        if exp_time == -1:
            return default
        if exp_time is not None and time() > exp_time:
            self.erase(key)
            return default
        return value

    def set(self, key, value, expiration: Optional[Union[datetime, timedelta]] = None):
        """Set a value with optional expiration.

        Args:
            expiration: datetime = deadline, timedelta = TTL, None = value will never expire.
        """
        if expiration is None:
            exp_time = None
        elif isinstance(expiration, datetime):
            exp_time = int(expiration.timestamp())
        elif isinstance(expiration, timedelta):
            exp_time = int(time() + expiration.total_seconds())
        else:
            raise TypeError('Invalid argument type')

        super().set(key, (value, exp_time))
